## Copyright (C) 2000  Leopoldo Cerbaro
##
## This file is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2, or (at your option)
## any later version.
##
## This file is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with Octave; see the file COPYING.  If not, write to the Free
## Software Foundation, 59 Temple Place - Suite 330, Boston, MA
## 02111-1307, USA.

## usage: geomview_display(STRING, STRUCT1, STRUCT2, ...)
##
## Display the objects described by STRUCT1,...;
## STRING is an (optional) string passed to geomview, e.g. " -nopanels";
## STRUCT is the geometrical object: curve, surface, triang. mesh
## see geomview.doc

## Author: Leopoldo Cerbaro cerbaro@keycomm.it
## Created: January 2000

## source it!
#
# this is not a function file

1;

function init_geomv_file()
#  tmp_in_file = tmpnam()
#  [fid, msg] = fopen(tmp_in_file, "w");
global fid;
  [fid, msg] = fopen("go.gv", "w");
  if (fid == -1) error("while opening gv file %s:\n\t%s", tmp_in_file, msg); endif
  fprintf ( fid, "(progn\n(geometry \"oct\" { = LIST {\n");
endfunction

function end_geomv_file(Plot)
  global fid;
  ## print common stuff
  lbox = max([Plot.xmax-Plot.xmin,Plot.ymax-Plot.ymin,Plot.zmax-Plot.zmin]);
  xcam = Plot.xmin + lbox * 1.4;
  ycam = Plot.ymin + lbox * 1.4;
  zcam = Plot.zmin + lbox * 1.4;
  lbox = lbox * 0.2; #now is used only for the font

fprintf ( fid, strcat(
"  { # the ranges \n",
"  appearance {linewidth 3}\n",
"  VECT\n",
"   8 19 3\n",
"   2 2 2 2 2 3 2 4 \n",
"   1 0 0 1 0 0 1 0\n\n",
sprintf("   %g 0 0 %g 0 0\n", 1.*Plot.xmin, 1.*Plot.xmax ),
sprintf("   %g %g %g %g %g %g\n", 
        1.*Plot.xmax , -.05*lbox, .05*lbox, 1.*Plot.xmax, .05*lbox, .25*lbox),
sprintf("   %g %g %g %g %g %g\n", 
        1.*Plot.xmax ,  .05*lbox, .05*lbox, 1.*Plot.xmax,-.05*lbox, .25*lbox),
sprintf("   0 %g 0  0 %g 0\n", 1.*Plot.ymin, 1.*Plot.ymax ),
sprintf("   %g %g %g %g %g %g\n", 
        -.05*lbox, 1.*Plot.ymax ,  .25*lbox, 0, 1.*Plot.ymax, .15*lbox),
sprintf("   %g %g %g %g %g %g %g %g %g\n",
 .05*lbox, 1.*Plot.ymax , .25*lbox, 0, 1.*Plot.ymax, .15*lbox, 0, 1.*Plot.ymax, .05*lbox ),
sprintf("   0 0 %g   0 0 %g\n", 1.*Plot.zmin, 1.*Plot.zmax ),
sprintf("   %g %g %g %g %g %g %g %g %g %g %g %g\n", 
 -.05*lbox, .25*lbox, 1.*Plot.zmax,  0.05*lbox, .25*lbox, 1.*Plot.zmax,
 -.05*lbox, .05*lbox, 1.*Plot.zmax,  0.05*lbox, .05*lbox, 1.*Plot.zmax ),
"   1 0 0 1\n",
"   0 1 0 1\n",
"   0 .6 1 1\n",
"  }\n" ));
if (( Plot.xmin > 0)||(Plot.xmax < 0) )
  fprintf ( fid, "  { # x-axes \n  appearance {linewidth 1}\n");
  fprintf ( fid, "  VECT\n   1 2 1\n   2 \n   1 \n\n" );
  if (abs(Plot.xmin)<abs(Plot.xmax)) 
    fprintf(fid, "   0 0 0 %g 0 0\n", Plot.xmin); 
  else 
    fprintf(fid, "   0 0 0 %g 0 0\n", Plot.xmax); 
  endif;
  fprintf (fid,"   0.6 0.  0.  1\n}\n" );
endif
if (( Plot.ymin > 0)||(Plot.ymax < 0) )
  fprintf (fid, "  { # y-axes \n  appearance {linewidth 1}\n");
  fprintf (fid, "  VECT\n   1 2 1\n   2 \n   1 \n\n" );
  if (abs(Plot.xmin)<abs(Plot.xmax)) 
    fprintf(fid, "   0 0 0 0 %g 0\n", Plot.ymin); 
  else 
    fprintf(fid, "   0 0 0 0 %g 0\n", Plot.ymax); 
  endif;
  fprintf (fid,"   0.  0.6 0.  1\n}\n" );
endif
if (( Plot.zmin > 0)||(Plot.zmax < 0) )
  fprintf (fid, "  { # z-axes \n  appearance {linewidth 1}\n");
  fprintf (fid, "  VECT\n   1 2 1\n   2 \n   1 \n\n" );
  if (abs(Plot.zmin)<abs(Plot.zmax)) 
    fprintf(fid, "   0 0 0 0  0 %g\n", Plot.zmin); 
  else 
    fprintf(fid, "   0 0 0 0  0 %g\n", Plot.zmax); 
  endif;
  fprintf (fid,"   0. 0.4 0.6 1\n}\n" );
endif
fprintf ( fid, strcat(
"} # end LIST \n", # end LIST
"}\n",
") # end geometry \"oct\"\n",
"(bbox-draw \"oct\" off)\n",
"(normalization \"oct\" none)\n",
"(camera \"Camera\" camera {\n",
"camtoworld transform {\n",
"             -0.707        0.707         0             0\n",
"             -0.707       -0.707         0             0\n",
"             0             0             1             0\n" ) );
fprintf ( fid,"     %g      %g      %g    1 \n}\n", xcam,ycam,zcam);
fprintf ( fid, strcat(
"	perspective 1  hyperbolic 0  stereo 0\n",
#"	fov 40\n",
"	frameaspect 1\n",
"}\n",
") # camera \"Camera\" ...\n",
#"(backcolor \"Camera\" 0.035414 0.114306 0.140603)\n",
"# end camera Camera\n",
"(look  oct Camera )\n",
"(ui-target \"World\")\n",
")\n" ) );
  fflush (fid);
  fclose(fid);
  command_string = sprintf('geomview -opengl %s %s',Plot.geomv_opts,"go.gv")
  system (command_string, 'sync');
endfunction

function write_geomv_obj(SS)
  global fid;
## print the objects ( curve, surface, parametric_surface, triang. mesh)
if ( strcmp(SS.type,"curve") )  # SS.X, SS.Y, SS.Z vector
  fprintf ( fid, "{ # curve\n  %s\n", SS.appearance );
  colors = (struct_contains(SS,"colorfunction"))*(SS.ncol-1)+1;#colors? NV or 1
  fprintf ( fid, "  VECT\n   1 %d %d\n   %d \n   %d\n\n", 
  	SS.ncol, colors, SS.ncol, colors );
  for i= 1:SS.ncol
    fprintf(fid, " %g %g %g \n", SS.XYZ(i,1), SS.XYZ(i,2), SS.XYZ(i,3) ); 
  endfor
  if (struct_contains(SS,"colorfunction"))
    if (SS.xmax != SS.xmin) xx = 1./(SS.xmax-SS.xmin); else xx = 0.; endif
    if (SS.ymax != SS.ymin) yy = 1./(SS.ymax-SS.ymin); else yy = 0.; endif   
    if (SS.zmax != SS.zmin) zz = 1./(SS.zmax-SS.zmin); else zz = 0.; endif
    for k=1:SS.ncol
      x = (SS.XYZ(k,1)-SS.xmin)*xx;
      y = (SS.XYZ(k,2)-SS.ymin)*yy;
      z = (SS.XYZ(k,3)-SS.zmin)*zz;
	    [r,g,b,a] = feval(SS.colorfunction, x, y, z, k);
      fprintf ( fid, "  %f  %f  %f %f\n", r, g, b, a);
    endfor
  else
    fprintf (fid,"   0.6 0.6  0.8  1\n" );
  endif
    fprintf (fid,"} # end curve\n" );
elseif (strcmp(SS.type,"tmesh") )      # -----------  triang.  mesh
  if ( struct_contains(SS,"colorfunction") )
    SS.vertexcolor="C";
  else
    SS.vertexcolor="";
  endif
  fprintf ( fid, "{\n %s\n %sOFF\n", SS.appearance, SS.vertexcolor );
  nvertices = SS.NV; nfaces = SS.NT;  nedges = SS.NV+SS.NT;
  fprintf ( fid, "%d %d %d\n", nvertices, nfaces, nedges);
  if (strcmp(SS.vertexcolor,"C") )
  	if (SS.xmax != SS.xmin) xx = 1./(SS.xmax-SS.xmin); else xx = 0.; endif
  	if (SS.ymax != SS.ymin) yy = 1./(SS.ymax-SS.ymin); else yy = 0.; endif   
 		if (SS.zmax != SS.zmin) zz = 1./(SS.zmax-SS.zmin); else zz = 0.; endif
    for i = 1:SS.NV
      x = (SS.xy(i,1) - SS.xmin)*xx; # normaliz. coord.
      y = (SS.xy(i,2) - SS.ymin)*yy;
      z = (SS.z(i)-SS.zmin)*zz;
      [r,g,b,a] = feval(SS.colorfunction, x, y, z, i);
      fprintf ( fid, "  %f  %f  %f  %f  %f  %f %f\n", 
               SS.xy(i,1), SS.xy(i,2), SS.z(i), r, g, b, a);
    endfor
  else # no color
    for i = 1:SS.NV
    fprintf ( fid, "  %f  %f  %f\n", SS.xy(i,1), SS.xy(i,2), SS.z(i));
    endfor
  endif
  for i=1:SS.NT
      fprintf ( fid, "3  %d %d %d\n", SS.tr(i,1), SS.tr(i,2), SS.tr(i,3));
  endfor
  fprintf ( fid, "}\n")
elseif (index(SS.type,"surface") ) 
  fprintf ( fid, "{\n %s\n %sOFF\n", SS.appearance, SS.vertexcolor );
  nvertices = SS.nrow*SS.ncol; nfaces = (SS.nrow-1)*(SS.ncol-1);
  nedges = SS.nrow*(SS.ncol-1)+SS.ncol*(SS.nrow-1);
  fprintf ( fid, "%d %d %d\n", nvertices, nfaces, nedges);
  if (strcmp(SS.type,"surface") ) # X, Y vector, Z matrix
  	if (strcmp(SS.vertexcolor,"C") )
	  	if (SS.xmax != SS.xmin) xx = 1./(SS.xmax-SS.xmin); else xx = 0.; endif
  		if (SS.ymax != SS.ymin) yy = 1./(SS.ymax-SS.ymin); else yy = 0.; endif
    	if (SS.zmax != SS.zmin) zz = 1./(SS.zmax-SS.zmin); else zz = 0.; endif
    	for i = 1:SS.nrow
      	for j = 1:SS.ncol
        	x = (SS.X(j) - SS.xmin)*xx; # normal. coord.
        	y = (SS.Y(i) - SS.ymin)*yy;
        	z = (SS.Z(i,j)-SS.zmin)*zz;
        	[r,g,b,a] = feval(SS.colorfunction, x, y, z );
        	fprintf ( fid, "  %f  %f  %f   %f  %f  %f %f\n", 
          	        SS.X(j), SS.Y(i), SS.Z(i,j),r, g, b, a);
      	endfor
    	endfor
  	else  # no color
    	for i = 1:SS.nrow
      	for j = 1:SS.ncol
        	fprintf ( fid, "  %f  %f  %f\n", SS.X(j), SS.Y(i), SS.Z(i,j));
      	endfor
    	endfor
  	endif
  else # paramametric_surface X, Y, Z are matrices
    if (strcmp(SS.vertexcolor,"C") )
      if (SS.xmax != SS.xmin) xx = 1./(SS.xmax-SS.xmin); else xx = 0.; endif
      if (SS.ymax != SS.ymin) yy = 1./(SS.ymax-SS.ymin); else yy = 0.; endif
      if (SS.zmax != SS.zmin) zz = 1./(SS.zmax-SS.zmin); else zz = 0.; endif
    	for i = 1:SS.nrow
      	for j = 1:SS.ncol
        	x = (SS.X(i,j)-SS.xmin)*xx; # normal. coord.
         	y = (SS.Y(i,j)-SS.ymin)*yy;
         	z = (SS.Z(i,j)-SS.zmin)*zz;
         	[r,g,b,a] = feval(SS.colorfunction, x, y, z );
          fprintf ( fid, "  %f  %f  %f %f  %f  %f %f\n", 
                   SS.X(i,j), SS.Y(i,j), SS.Z(i,j), r, g, b, a);
	      endfor
      endfor
    else
    	for i = 1:SS.nrow
      	for j = 1:SS.ncol
          fprintf ( fid, "  %f  %f  %f\n", SS.X(i,j), SS.Y(i,j), SS.Z(i,j) );
	      endfor
      endfor
    endif
  endif  
  for i=1:SS.nrow-1
    for j=1:SS.ncol-1
      i1 = (i - 1) * SS.ncol + j - 1;
      i2 = (i - 1) * SS.ncol + j;
      i3 = i * SS.ncol + j;
      i4 = i * SS.ncol + j - 1;
      fprintf ( fid, "4  %d %d %d %d\n", i1, i2, i3, i4);
    endfor
  endfor
  fprintf ( fid, "}\n");
	endif
endfunction

function geomview_display (...)
  init_geomv_file();
  Plot.geomv_opts ="";
  while (nargin--)
    SS = va_arg(); # current arguments
    if ( isstr(SS) )
      Plot.geomv_opts = SS;  # geomview options: -wpos ..., -nopanels, ...
      continue
    elseif ( !is_struct(SS) || !struct_contains(SS,"type") )
      disp("argument not a struct or type unknown; ...trying to continue");
      continue;  # process the next arg
    elseif ( strcmp(SS.type,"tmesh") )  # triang. mesh
      # mandatory args: 
      #   SS.xy vert. coord.; SS.z z at vert.; SS.tr vert. of triang from 0!
      if ( !(struct_contains(SS,"xy") && struct_contains(SS,"z") &&
             struct_contains(SS,"tr") ) )
        error("Triang. mesh: arguments required")
      endif
      SS.NV = rows(SS.xy);
      SS.NT = rows(SS.tr);
      SS.xmin = min(SS.xy(:,1)); SS.xmax = max(SS.xy(:,1));
      SS.ymin = min(SS.xy(:,2)); SS.ymax = max(SS.xy(:,2));
      SS.zmin = min(SS.z);       SS.zmax = max(SS.z);
      if ( !struct_contains(SS,"appearance") )
        SS.appearance = " appearance {shading smooth}";
      endif  
      if (struct_contains(Plot,"xmin") ) # and so Plot.ymin ... also
        Plot.xmin = min(Plot.xmin,SS.xmin); Plot.xmax = max(Plot.xmax,SS.xmax);
        Plot.ymin = min(Plot.ymin,SS.ymin); Plot.ymax = max(Plot.ymax,SS.ymax);
        Plot.zmin = min(Plot.zmin,SS.zmin); Plot.zmax = max(Plot.zmax,SS.zmax);
      else
        Plot.xmin = SS.xmin; Plot.xmax = SS.xmax;
        Plot.ymin = SS.ymin; Plot.ymax = SS.ymax;
        Plot.zmin = SS.zmin; Plot.zmax = SS.zmax;
      endif        
    elseif (strcmp(SS.type,"curve") )   # space curve
      # mandatory args: 
      #   SS.X X coords; SS.Y y coords; SS.Z z coords
      if ( !(struct_contains(SS, "XYZ") && is_matrix(SS.XYZ) && 
        		 columns(SS.XYZ) >= 3 ) ) 
        error("Curve: wrong argd")
      endif
        SS.ncol = rows (SS.XYZ);
        SS.xmin = min(SS.XYZ(:,1)); SS.xmax = max(SS.XYZ(:,1));
        SS.ymin = min(SS.XYZ(:,2)); SS.ymax = max(SS.XYZ(:,2));
        SS.zmin = min(SS.XYZ(:,3)); SS.zmax = max(SS.XYZ(:,3));
	      if ( !struct_contains(SS,"appearance") )
          SS.appearance = " appearance {linewidth 2}";    	  
        endif  
	      if (struct_contains(Plot,"xmin") ) # and so Plot.ymin ... also        
  	      Plot.xmin=min(Plot.xmin,SS.xmin); Plot.xmax=max(Plot.xmax,SS.xmax);
    	    Plot.ymin=min(Plot.ymin,SS.ymin); Plot.ymax=max(Plot.ymax,SS.ymax);
      	  Plot.zmin=min(Plot.zmin,SS.zmin); Plot.zmax=max(Plot.zmax,SS.zmax);
      	else
        	Plot.xmin = SS.xmin; Plot.xmax = SS.xmax;
        	Plot.ymin = SS.ymin; Plot.ymax = SS.ymax;
        	Plot.zmin = SS.zmin; Plot.zmax = SS.zmax;
      	endif
    elseif (strcmp(SS.type,"surface") )   # Explicit Surface
      # args: 
      # matrix SS.Z | (matrix SS.Z, vects SS.X,SS.Y) | matrix SS.X SS.Y SS.Z
      # vects SS.X SS.Y (mesh only)
      if ( !struct_contains(SS, "Z") ) 
        error("Surface: matrix for Z values required")
      else
        SS.ncol = columns(SS.Z); SS.nrow = rows(SS.Z);
        SS.zmin = min(min(SS.Z)); SS.zmax = max(max(SS.Z));
      endif
      if ( struct_contains(SS, "X") && struct_contains(SS, "Y") )
        if (is_vector(SS.X) && SS.ncol == length(SS.X) &&
            is_vector(SS.Y) && SS.nrow == length(SS.Y) )
          SS.xmin=min(SS.X); SS.xmax=max(SS.X);
          SS.ymin=min(SS.Y); SS.ymax=max(SS.Y);
        elseif (is_matrix(SS.X) && 
                SS.ncol == columns(SS.X) && SS.nrow == rows(SS.X) &&
                is_matrix(SS.Y) && 
                SS.ncol == columns(SS.Y) && SS.nrow == rows(SS.Y) )
          SS.type = "parametric_surface";   # Parametric Surface
          SS.xmin=min(min(SS.X)); SS.xmax=max(max(SS.X));
          SS.ymin=min(min(SS.Y)); SS.ymax=max(max(SS.Y));      
        else
          error("Surface: compatible vector or matrix for X or Y values required")
        endif
      else
        SS.X = 1:ncol; SS.xmin = 1; SS.xmax = ncol;
        SS.Y = 1:nrow; SS.ymin = 1; SS.ymax = nrow;
      endif
      if ( !struct_contains(SS,"appearance") )
        SS.appearance = " appearance {shading smooth}";
      endif  
      if ( struct_contains(SS, "colorfunction") )
        SS.vertexcolor="C";
      else  
        SS.vertexcolor="";
      endif
      if (struct_contains(Plot,"xmin") ) 
 	      Plot.xmin=min(Plot.xmin,SS.xmin); Plot.xmax=max(Plot.xmax,SS.xmax);
   	    Plot.ymin=min(Plot.ymin,SS.ymin); Plot.ymax=max(Plot.ymax,SS.ymax);
     	  Plot.zmin=min(Plot.zmin,SS.zmin); Plot.zmax=max(Plot.zmax,SS.zmax);
     	else
       	Plot.xmin = SS.xmin; Plot.xmax = SS.xmax;
       	Plot.ymin = SS.ymin; Plot.ymax = SS.ymax;
       	Plot.zmin = SS.zmin; Plot.zmax = SS.zmax;
     	endif
    else
      error("Unknown type")
    endif
    write_geomv_obj(SS);
  endwhile
  end_geomv_file(Plot);
endfunction
